# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

from dataclasses import dataclass
from typing import NewType

import scipp as sc

Ltotal = NewType("Ltotal", sc.Variable)
"""
Total length of the flight path from the source to the detector.
"""


@dataclass
class SimulationResults:
    """
    Results of a time-of-flight simulation used to create a lookup table.

    The results (apart from ``distance``) should be flat lists (1d arrays) of length N
    where N is the number of neutrons, containing the properties of the neutrons in the
    simulation.

    Parameters
    ----------
    time_of_arrival:
        Time of arrival of the neutrons at the position where the events were recorded
        (1d array of size N).
    speed:
        Speed of the neutrons, typically derived from the wavelength of the neutrons
        (1d array of size N).
    wavelength:
        Wavelength of the neutrons (1d array of size N).
    weight:
        Weight/probability of the neutrons (1d array of size N).
    distance:
        Distance from the source to the position where the events were recorded
        (single value; we assume all neutrons were recorded at the same position).
        For a ``tof`` simulation, this is just the position of the detector where the
        events are recorded. For a ``McStas`` simulation, this is the distance between
        the source and the event monitor.
    """

    time_of_arrival: sc.Variable
    speed: sc.Variable
    wavelength: sc.Variable
    weight: sc.Variable
    distance: sc.Variable


@dataclass
class FastestNeutron:
    """
    Properties of the fastest neutron in the simulation results.
    """

    time_of_arrival: sc.Variable
    speed: sc.Variable
    distance: sc.Variable


LtotalRange = NewType("LtotalRange", tuple[sc.Variable, sc.Variable])
"""
Range (min, max) of the total length of the flight path from the source to the detector.
This is used to create the lookup table to compute the neutron time-of-flight.
Note that the resulting table will extend slightly beyond this range, as the supplied
range is not necessarily a multiple of the distance resolution.

Note also that the range of total flight paths is supplied manually to the workflow
instead of being read from the input data, as it allows us to compute the expensive part
of the workflow in advance (the lookup table) and does not need to be repeated for each
run, or for new data coming in in the case of live data collection.
"""

DistanceResolution = NewType("DistanceResolution", sc.Variable)
"""
Step size of the distance axis in the lookup table.
Should be a single scalar value with a unit of length.
This is typically of the order of 1-10 cm.
"""

TimeOfArrivalResolution = NewType("TimeOfArrivalResolution", int | sc.Variable)
"""
Resolution of the time of arrival axis in the lookup table.
Can be an integer (number of bins) or a sc.Variable (bin width).
"""

TimeOfFlightLookupTable = NewType("TimeOfFlightLookupTable", sc.DataArray)
"""
Lookup table giving time-of-flight as a function of distance and time of arrival.
"""

MaskedTimeOfFlightLookupTable = NewType("MaskedTimeOfFlightLookupTable", sc.DataArray)
"""
Lookup table giving time-of-flight as a function of distance and time of arrival, with
regions of large uncertainty masked out.
"""

LookupTableRelativeErrorThreshold = NewType("LookupTableRelativeErrorThreshold", float)

FramePeriod = NewType("FramePeriod", sc.Variable)
"""
The period of a frame, a (small) integer multiple of the source period.
"""

UnwrappedTimeOfArrival = NewType("UnwrappedTimeOfArrival", sc.Variable)
"""
Time of arrival of the neutron at the detector, unwrapped at the pulse period.
"""

PivotTimeAtDetector = NewType("PivotTimeAtDetector", sc.Variable)
"""
Pivot time at the detector, i.e., the time of the start of the frame at the detector.
"""

UnwrappedTimeOfArrivalMinusPivotTime = NewType(
    "UnwrappedTimeOfArrivalMinusPivotTime", sc.Variable
)
"""
Time of arrival of the neutron at the detector, unwrapped at the pulse period, minus
the start time of the frame.
"""

TimeOfArrivalMinusPivotTimeModuloPeriod = NewType(
    "TimeOfArrivalMinusPivotTimeModuloPeriod", sc.Variable
)
"""
Time of arrival of the neutron at the detector minus the start time of the frame,
modulo the frame period.
"""

FrameFoldedTimeOfArrival = NewType("FrameFoldedTimeOfArrival", sc.Variable)


PulsePeriod = NewType("PulsePeriod", sc.Variable)
"""
Period of the source pulses, i.e., time between consecutive pulse starts.
"""

PulseStride = NewType("PulseStride", int)
"""
Stride of used pulses. Usually 1, but may be a small integer when pulse-skipping.
"""

PulseStrideOffset = NewType("PulseStrideOffset", int)
"""
When pulse-skipping, the offset of the first pulse in the stride.
"""

RawData = NewType("RawData", sc.DataArray)
"""
Raw detector data loaded from a NeXus file, e.g., NXdetector containing NXevent_data.
"""

TofData = NewType("TofData", sc.DataArray)
"""
Detector data with time-of-flight coordinate.
"""

ResampledTofData = NewType("ResampledTofData", sc.DataArray)
"""
Histogrammed detector data with time-of-flight coordinate, that has been resampled.

Histogrammed data that has been converted to `tof` will typically have
unsorted bin edges (due to either wrapping of `time_of_flight` or wavelength
overlap between subframes).
We thus resample the data to ensure that the bin edges are sorted.
It makes use of the ``to_events`` helper which generates a number of events in each
bin with a uniform distribution. The new events are then histogrammed using a set of
sorted bin edges to yield a new histogram with sorted bin edges.

WARNING:
This function is highly experimental, has limitations and should be used with
caution. It is a workaround to the issue that rebinning data with unsorted bin
edges is not supported in scipp.
"""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import pytest
import scipp as sc

from ess.reduce.time_of_flight.to_events import to_events


def test_to_events_1d():
    table = sc.data.table_xyz(1000)
    hist = table.hist(x=20)
    events = to_events(hist, "event")
    assert "x" not in events.dims
    result = events.hist(x=hist.coords["x"])
    assert sc.identical(hist.coords["x"], result.coords["x"])
    assert sc.allclose(hist.data, result.data)


def test_to_events_1d_with_group_coord():
    table = sc.data.table_xyz(1000, coord_max=10)
    table.coords["l"] = table.coords["x"].to(dtype=int)
    hist = table.group("l").hist(x=20)
    events = to_events(hist, "event")
    assert "x" not in events.dims
    assert "l" in events.dims
    result = events.hist(x=hist.coords["x"])
    assert sc.identical(hist.coords["x"], result.coords["x"])
    assert sc.identical(hist.coords["l"], result.coords["l"])
    assert sc.allclose(hist.data, result.data)


def test_to_events_2d():
    table = sc.data.table_xyz(1000)
    hist = table.hist(y=20, x=10)
    events = to_events(hist, "event")
    assert "x" not in events.dims
    assert "y" not in events.dims
    result = events.hist(y=hist.coords["y"], x=hist.coords["x"])
    assert sc.identical(hist.coords["x"], result.coords["x"])
    assert sc.identical(hist.coords["y"], result.coords["y"])
    assert sc.allclose(hist.data, result.data)


def test_to_events_2d_with_group_coord():
    table = sc.data.table_xyz(1000, coord_max=10)
    table.coords["l"] = table.coords["x"].to(dtype=int)
    hist = table.group("l").hist(y=20, x=10)
    events = to_events(hist, "event")
    assert "x" not in events.dims
    assert "y" not in events.dims
    assert "l" in events.dims
    result = events.hist(y=hist.coords["y"], x=hist.coords["x"])
    assert sc.identical(hist.coords["x"], result.coords["x"])
    assert sc.identical(hist.coords["y"], result.coords["y"])
    assert sc.identical(hist.coords["l"], result.coords["l"])
    assert sc.allclose(hist.data, result.data)


def test_to_events_binned_data_input_raises():
    table = sc.data.table_xyz(1000)
    binned = table.bin(x=20)
    with pytest.raises(ValueError, match="Cannot convert a binned DataArray to events"):
        _ = to_events(binned, "event")


def test_to_events_mask_on_midpoints_dim():
    table = sc.data.table_xyz(1000, coord_max=10)
    table.coords["l"] = table.coords["y"].to(dtype=int)

    hist = table.group("l").hist(x=6)
    hist.masks["m"] = hist.coords["l"] == sc.scalar(3.0, unit="m")
    events = to_events(hist, "event")
    result = events.hist(x=hist.coords["x"])
    assert sc.identical(hist.coords["x"], result.coords["x"])
    assert sc.allclose(hist.data, result.data)
    assert sc.identical(hist.masks["m"], result.masks["m"])


def test_to_events_mask_on_binedge_dim():
    table = sc.data.table_xyz(1000, coord_max=10)
    table.coords["l"] = table.coords["x"].to(dtype=int)

    hist = table.group("l").hist(x=6)
    hist.masks["m"] = sc.array(
        dims=["x"], values=[False, False, True, True, False, False]
    )
    events = to_events(hist, "event")
    result = events.hist(x=hist.coords["x"])
    assert sc.identical(hist.coords["x"], result.coords["x"])
    assert sc.isclose(hist.sum().data, result.sum().data)
    assert "m" not in result.masks
    assert result["x", 2:4].data.sum() == sc.scalar(0.0, unit=table.unit)


def test_to_events_two_masks():
    table = sc.data.table_xyz(1000, coord_max=10)
    table.coords["l"] = table.coords["x"].to(dtype=int)

    hist = table.group("l").hist(x=6)
    hist.masks["m1"] = sc.array(
        dims=["x"], values=[False, False, True, True, False, False]
    )
    hist.masks["m2"] = hist.coords["l"] == sc.scalar(3.0, unit="m")
    events = to_events(hist, "event")
    result = events.hist(x=hist.coords["x"])
    assert sc.identical(hist.coords["x"], result.coords["x"])
    assert sc.isclose(hist.sum().data, result.sum().data)
    assert "m1" not in result.masks
    assert sc.identical(hist.masks["m2"], result.masks["m2"])
    assert result["x", 2:4].data.sum() == sc.scalar(0.0, unit=table.unit)


def test_to_events_1d_unsorted_bin_edges():
    table = sc.data.table_xyz(1000)
    hist = table.hist(x=10)
    hist.coords["x"].values = hist.coords["x"].values[
        [0, 1, 2, 3, 5, 4, 6, 7, 8, 9, 10]
    ]
    events = to_events(hist, "event")
    assert "x" not in events.dims
    result = events.hist(x=sc.sort(hist.coords["x"], "x"))
    assert sc.allclose(hist.data[:3], result.data[:3])
    assert sc.allclose(hist.data[6:], result.data[6:])
    # The data in the middle gets re-ordered, but the sum should still be the same
    assert sc.isclose(hist.data[3:6].sum(), result.data[3:6].sum())



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401

import importlib.metadata

from . import nexus, uncertainty, time_of_flight

try:
    __version__ = importlib.metadata.version("essreduce")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib

__all__ = ["nexus", "uncertainty", "time_of_flight"]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import pooch


class Registry:
    def __init__(self, instrument: str, files: dict[str, str], version: str):
        self._registry = pooch.create(
            path=pooch.os_cache(f'ess/{instrument}'),
            env=f'ESS_{instrument.upper()}_DATA_DIR',
            base_url=f'https://public.esss.dk/groups/scipp/ess/{instrument}/'
            + '{version}/',
            version=version,
            registry=files,
        )

    def __contains__(self, key):
        return key in self._registry.registry

    def get_path(self, name: str, unzip: bool = False) -> str:
        """
        Get the path to a file in the registry.

        Parameters
        ----------
        name:
            Name of the file to get the path for.
        unzip:
            If `True`, unzip the file before returning the path.
        """
        return self._registry.fetch(name, processor=pooch.Unzip() if unzip else None)


_bifrost_registry = Registry(
    instrument='bifrost',
    files={
        "BIFROST_20240914T053723.h5": "md5:0f2fa5c9a851f8e3a4fa61defaa3752e",
    },
    version='1',
)


_loki_registry = Registry(
    instrument='loki',
    files={
        # Files from LoKI@Larmor detector test experiment
        #
        # Background run 1 (no sample, sample holder/can only, no transmission monitor)
        '60248-2022-02-28_2215.nxs': 'md5:d9f17b95274a0fc6468df7e39df5bf03',
        # Sample run 1 (sample + sample holder/can, no transmission monitor in beam)
        '60250-2022-02-28_2215.nxs': 'md5:6a519ceaacbae702a6d08241e86799b1',
        # Sample run 2 (sample + sample holder/can, no transmission monitor in beam)
        '60339-2022-02-28_2215.nxs': 'md5:03c86f6389566326bb0cbbd80b8f8c4f',
        # Background transmission run (sample holder/can + transmission monitor)
        '60392-2022-02-28_2215.nxs': 'md5:9ecc1a9a2c05a880144afb299fc11042',
        # Background run 2 (no sample, sample holder/can only, no transmission monitor)
        '60393-2022-02-28_2215.nxs': 'md5:bf550d0ba29931f11b7450144f658652',
        # Sample transmission run (sample + sample holder/can + transmission monitor)
        '60394-2022-02-28_2215.nxs': 'md5:c40f38a62337d86957af925296c4c615',
        # Analytical model for the I(Q) of the Poly-Gauss sample
        'PolyGauss_I0-50_Rg-60.h5': 'md5:f5d60d9c2286cb197b8cd4dc82db3d7e',
        # XML file for the pixel mask
        'mask_new_July2022.xml': 'md5:421b6dc9db74126ffbc5d88164d017b0',
    },
    version='2',
)


def bifrost_simulated_elastic() -> str:
    """McStas simulation with elastic incoherent scattering + phonon."""
    return _bifrost_registry.get_path('BIFROST_20240914T053723.h5')


def loki_tutorial_sample_run_60250() -> str:
    """Sample run with sample and sample holder/can, no transmission monitor in beam."""
    return _loki_registry.get_path('60250-2022-02-28_2215.nxs')


def loki_tutorial_sample_run_60339() -> str:
    """Sample run with sample and sample holder/can, no transmission monitor in beam."""
    return _loki_registry.get_path('60339-2022-02-28_2215.nxs')


def loki_tutorial_background_run_60248() -> str:
    """Background run with sample holder/can only, no transmission monitor."""
    return _loki_registry.get_path('60248-2022-02-28_2215.nxs')


def loki_tutorial_background_run_60393() -> str:
    """Background run with sample holder/can only, no transmission monitor."""
    return _loki_registry.get_path('60393-2022-02-28_2215.nxs')


def loki_tutorial_sample_transmission_run() -> str:
    """Sample transmission run (sample + sample holder/can + transmission monitor)."""
    return _loki_registry.get_path('60394-2022-02-28_2215.nxs')
